"""
slurminade allows to distribute function calls to slurm using decorators.

.. code-block:: python

    import slurminade

    slurminade.update_default_configuration(partition="alg")  # global options for slurm

    # If no slurm environment is found, the functions are called directly to make scripts
    # compatible with any environment.
    # You can enforce slurm with `slurminade.set_dispatcher(slurminade.SlurmDispatcher())`


    # use this decorator to make a function distributable with slurm
    @slurminade.slurmify(
        constraint="alggen02"
    )  # function specific options can be specified
    def prepare():
        print("Prepare")


    @slurminade.slurmify()
    def f(foobar):
        print(f"f({foobar})")


    @slurminade.slurmify()
    def clean_up():
        print("Clean up")


    if __name__ == "__main__":
        jid = prepare.distribute()

        with slurminade.Batch(max_size=20) as batch:  # automatically bundles up to 20 tasks
            # run 100x f after `prepare` has finished
            for i in range(100):
                f.wait_for(jid).distribute(i)

            # clean up after the previous jobs have finished
            jids = batch.flush()
            clean_up.wait_for(jids).distribute()


Project structure:
- batch.py: Contains code for bundling tasks, so we don't spam slurm with too many.
- conf.py: Contains code for managing the configuration of slurm.
- dispatcher.py: Contains code for actually dispatching tasks to slurm.
- execute.py: Contains code to execute the task on the slurm node.
- function.py: Contains the code for making a function slurm-compatible.
- function_map.py: Saves all the slurified functions.
- guard.py: Contains code to prevent you accidentally DDoSing your infrastructure.
- options.py: Contains a simple data structure to save slurm options.
"""

# flake8: noqa F401
from .function import slurmify
from .conf import update_default_configuration, set_default_configuration
from .guard import (
    set_dispatch_limit,
    allow_recursive_distribution,
    disable_warning_on_repeated_flushes,
)
from .batch import Batch
from .dispatcher import (
    srun,
    sbatch,
    SlurmDispatcher,
    set_dispatcher,
    get_dispatcher,
    TestDispatcher,
    SubprocessDispatcher,
)
from .function_map import set_entry_point
from .node_setup import node_setup

__all__ = [
    "slurmify",
    "update_default_configuration",
    "set_default_configuration",
    "set_dispatch_limit",
    "allow_recursive_distribution",
    "disable_warning_on_repeated_flushes",
    "Batch",
    "srun",
    "sbatch",
    "SlurmDispatcher",
    "set_dispatcher",
    "get_dispatcher",
    "TestDispatcher",
    "SubprocessDispatcher",
    "set_entry_point",
    "node_setup",
]

# set default logging
import logging
import sys

# Set up the root logger to print to stdout by default
logging.basicConfig(level=logging.INFO, stream=sys.stdout)



import unittest
from pathlib import Path

import slurminade

test_file_path = Path("./f_test_file.txt")


@slurminade.slurmify()
def f(s):
    with open(test_file_path, "w") as file:
        file.write(s)


class TestCreateCommand(unittest.TestCase):
    def test_dispatch_with_temp_file(self):
        slurminade.set_entry_point(__file__)
        if test_file_path.exists():
            test_file_path.unlink()
        dispatcher = slurminade.SubprocessDispatcher()
        dispatcher.max_arg_length = 1
        slurminade.set_dispatcher(dispatcher)
        s = "test"
        f.distribute(s)
        assert test_file_path.is_file()
        with open(test_file_path) as file:
            assert file.readline() == s
        if test_file_path.exists():  # delete the file
            test_file_path.unlink()



import slurminade

slurminade.set_dispatch_limit(10)  # allow maximally 10 dispatches
slurminade.update_default_configuration(partition="alg", constraint="alggen02")
# slurminade.set_dispatcher(slurminade.SlurmDispatcher())  # enforce slurm


@slurminade.slurmify()
def prepare():
    print("Prepare")


@slurminade.slurmify()
def f(foobar):
    print(f"f({foobar})")


@slurminade.slurmify()
def clean_up():
    print("Clean up")


if __name__ == "__main__":
    jid = prepare.distribute()

    with slurminade.Batch(max_size=20) as batch:  # automatically bundles up to 20 tasks
        # run 10x f after prepare has finished
        for i in range(100):
            f.wait_for(jid).distribute(i)

        # clean up after the previous jobs have finished
        jids = batch.flush()
        clean_up.wait_for(jids).distribute()



import typing


class SlurmOptions(dict):
    """
    Primarily just a wrapper to allow using the options in a dict as key.
    Necessary for batching function calls, because only function calls with the
    same options can be bundled.
    """

    def _items(self):
        for k, v in self.items():
            if isinstance(v, dict):
                v = SlurmOptions(**v)
            yield k, v

    def __hash__(self):
        return hash(tuple(sorted(hash((k, v)) for k, v in self._items())))

    def __eq__(self, other):
        if not isinstance(other, SlurmOptions):
            return False
        return hash(self) == hash(other)

    def as_dict(self) -> typing.Dict:
        return dict(self._items())



import unittest

import slurminade
from slurminade.guard import TooManyDispatchesError, _DispatchGuard, set_dispatch_limit


@slurminade.slurmify()
def f():
    pass


class TestDispatchGuard(unittest.TestCase):
    def test_simple(self):
        slurminade.set_entry_point(__file__)
        dg = _DispatchGuard(3)
        dg()
        dg()
        dg()
        self.assertRaises(TooManyDispatchesError, dg)

    def test_dispatch_limit(self):
        slurminade.set_entry_point(__file__)
        set_dispatch_limit(3)
        f.distribute()
        f.distribute()
        f.distribute()
        self.assertRaises(TooManyDispatchesError, f.distribute)

    def test_dispatch_limit_batch(self):
        slurminade.set_entry_point(__file__)
        set_dispatch_limit(2)
        with slurminade.Batch(max_size=2):
            for _ in range(4):
                f.distribute()
        self.assertRaises(TooManyDispatchesError, f.distribute)



import inspect
import typing
from .guard import on_slurm_node

def node_setup(func: typing.Callable):
    """
    Decorator: Call this function on the node before running any function calls.
    """
    if on_slurm_node():
        func()
    else:
        # check if the function has no arguments
        sig = inspect.signature(func)
        if sig.parameters:
            msg = "The node setup function must not have any arguments."
            raise ValueError(msg)
    return func



import os.path
import unittest

import slurminade

f_file = "./f_test_file.txt"
g_file = "./g_test_file.txt"


@slurminade.slurmify()
def f():
    with open(f_file, "w") as file:
        file.write("test")


def delete_f():
    if os.path.exists(f_file):
        os.remove(f_file)


@slurminade.slurmify()
def g(x, y):
    with open(g_file, "w") as file:
        file.write(f"{x}:{y}")


def delete_g():
    if os.path.exists(g_file):
        os.remove(g_file)


def get_file_name():
    filename = os.getenv("PYTEST_CURRENT_TEST")  # doesn't seem to work always
    filename = filename.split("::")[0] if filename else __file__  # workaround
    if not os.path.exists(filename):  # sometimes the test folder gets duplicated.
        filename = "/".join(filename.split("/")[1:])
    return os.path.abspath(filename)


class TestSubprocess(unittest.TestCase):
    def test_1(self):
        slurminade.set_dispatcher(slurminade.SubprocessDispatcher())
        slurminade.set_dispatch_limit(100)
        slurminade.set_entry_point(get_file_name())

        delete_f()
        f.distribute()
        assert os.path.exists(f_file)
        delete_f()

    def test_2(self):
        slurminade.set_dispatcher(slurminade.SubprocessDispatcher())
        slurminade.set_entry_point(get_file_name())
        slurminade.set_dispatch_limit(100)

        delete_g()
        g.distribute(x="a", y=2)
        assert os.path.exists(g_file)
        with open(g_file) as file:
            assert file.readline() == "a:2"
        delete_g()




import slurminade
from pytest import raises

def test_dispatch_limit_batch():
    @slurminade.slurmify()
    def f():
        pass

    slurminade.set_entry_point(__file__)

    with raises(KeyError):
        f.distribute()



import slurminade
from pathlib import Path

test_file_path = Path("./f_test_file.txt")

@slurminade.node_setup
def f():
    with open(test_file_path, "w") as file:
        file.write('node_setup')

@slurminade.slurmify
def nil():
    pass

def test_node_setup():
    slurminade.set_entry_point(__file__)
    if test_file_path.exists():
        test_file_path.unlink()
    dispatcher = slurminade.SubprocessDispatcher()
    slurminade.set_dispatcher(dispatcher)
    slurminade.set_dispatch_limit(100)
    nil.distribute()
    with open(test_file_path) as file:
        assert file.readline() == 'node_setup'
    if test_file_path.exists():  # delete the file
        test_file_path.unlink()


import inspect
import typing

from .dispatcher import FunctionCall, dispatch, get_dispatcher
from .function_map import FunctionMap
from .guard import guard_recursive_distribution
from .options import SlurmOptions


class SlurmFunction:
    """
    A wrapper around a function that allows it to be distributed to slurm.

    .. code-block:: python

        @slurmify(function specific slurm options)
        def f(foobar):
            print(foobar)

        if __name__=="__main__":
            assert isinstance(f, SlurmFunction), "f has become a SlurmFunction"
            jid = f.distribute("hello")
            f.wait_for(jid).distribute("bye")

    """

    def __init__(
        self, special_slurm_opts: typing.Dict, func: typing.Callable, func_id: str
    ):
        self.special_slurm_opts = SlurmOptions(**special_slurm_opts)
        self.func = func
        self.func_id = func_id

    def update_options(self, conf: typing.Dict[str, typing.Any]):
        self.special_slurm_opts.update(conf)

    def _add_dependencies(self, job_ids, method: str = "afterany"):
        opt = f"{method}:" + ":".join(str(jid) for jid in job_ids)
        if "dependency" in self.special_slurm_opts:
            # There are already dependencies. Trying to extend them.
            if isinstance(self.special_slurm_opts["dependency"], dict):
                dependecy_dict = self.special_slurm_opts["dependency"]
                if method in dependecy_dict:
                    dependecy_dict[method] += ":" + ":".join(
                        str(jid) for jid in job_ids
                    )
                else:
                    dependecy_dict[method] = ":".join(str(jid) for jid in job_ids)
            elif isinstance(self.special_slurm_opts["dependency"], str):
                self.special_slurm_opts["dependency"] += "," + opt
            else:
                # Could not extend dependencies because I have no idea what is going on.
                msg = "Key 'dependency' has unexpected type."
                raise RuntimeError(msg)
        else:
            self.special_slurm_opts["dependency"] = opt

    def wait_for(
        self, job_ids: typing.Union[int, typing.Iterable[int]], method: str = "afterany"
    ) -> "SlurmFunction":
        """
        Add a dependency to a distribution.
        `f_jid = f.wait_for(job_ids).distribute("hello")`
        f will only be executed once all jobs within job_ids have finished (or failed).
        Its own job id can then also be used as dependency for other distributions.
        Note that this method does not manipulate the original function but returns
        a new function object, so you have to use chaining.
        Not every dispatcher returns valid job ids on distribute. For example with
        `Batch`, you get the job ids with `job_ids = batch.flush()`.
        :param job_ids: A single job id or an iterable of job ids.
        :param method: 'after'|'afterany'|'afternotok'|'afterok'|'singleton'
        :return: Chainable slurm function object.
        """
        sfunc = SlurmFunction(self.special_slurm_opts, self.func, self.func_id)
        if isinstance(job_ids, int):
            job_ids = [job_ids]
        else:
            job_ids = list(job_ids)  # make sure it is a list
        if not job_ids and not get_dispatcher().is_sequential():
            msg = "Creating a dependency on an empty list of job ids."
            msg += " This is probably an error in your code."
            msg += " Maybe you are using `Batch` but flush outside of the `with` block?"
            raise RuntimeError(msg)
        if any(jid < 0 for jid in job_ids) and not get_dispatcher().is_sequential():
            msg = "Invalid job id. Not every dispatcher can directly return job ids, because it may not directly distribute them or doesn't distribute them at all."
            raise RuntimeError(msg)
        sfunc._add_dependencies(list(job_ids), method)
        return sfunc

    def _check(self, args, kwargs):
        """
        Check if the arguments match the function signature.
        """
        inspect.signature(self.func).bind(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        """
        Direct call of the original function.
        :param args: Positional arguments.
        :param kwargs: Keyword arguments.
        :return: The return value of the function.
        """
        return self.func(*args, **kwargs)

    def distribute(self, *args, **kwargs) -> int:
        """
        Try to distribute function call. If slurm is not available, a direct function
        call will be performed.
        `f.distribute("hello")`
        Call with function arguments.
        :return: Job id. Not necessarily valid (usually -1 in this case).
        """
        self._check(args, kwargs)
        guard_recursive_distribution()
        return dispatch(
            [FunctionCall(self.func_id, args, kwargs)], self.special_slurm_opts
        )

    def __str__(self) -> str:
        return self.func.__name__

    @staticmethod
    def call(func_id, *args, **kwargs):
        return FunctionMap.call(func_id, args, kwargs)


def slurmify(f=None, **args) -> typing.Callable[[typing.Callable], SlurmFunction]:
    """
    Decorator: Make a function distributable to slurm.
    Usage:

    .. code-block:: python

        @slurmify()
        def func(a, b):
            pass

    :param f: Function
    :param args: Special slurm options for this function.
    :return: A decorated function, callable with slurm.
    """

    if f:  # use default parameters
        func_id = FunctionMap.register(f)
        return SlurmFunction({}, f, func_id)
    else:

        def dec(func) -> SlurmFunction:
            func_id = FunctionMap.register(func)
            return SlurmFunction(args, func, func_id)

        return dec



"""
The dispatcher distribute function calls to slurm or the local machine.
It can be accessed with `get_dispatcher` and set with `set_dispatcher`.
This allows to change the behaviour of the distribution, e.g., we use it for batch:
Batch simply wraps the dispatcher by a buffered version.
"""
import abc
import json
import logging
import os
import shlex
import shutil
import subprocess
import sys
import typing
from tempfile import mkstemp

import simple_slurm

from .conf import _get_conf
from .function_map import FunctionMap, get_entry_point
from .guard import dispatch_guard
from .options import SlurmOptions

# MAX_ARG_STRLEN on a Linux system with PAGE_SIZE 4096 is 131072
DEFAULT_MAX_ARG_LENGTH = 100000


class FunctionCall:
    """
    A function call to be dispatched.
    """

    def __init__(self, func_id, args, kwargs):
        self.func_id = func_id  # the function id, as in FunctionMap
        self.args = args  # the positional arguments for the call
        self.kwargs = kwargs  # the keyword arguments for the call

    def to_json(self) -> typing.Dict:
        """
        Convert call to a json object that can be passed to slurm.
        :return: json object.
        """
        return {"func_id": self.func_id, "args": self.args, "kwargs": self.kwargs}

    def __str__(self) -> str:
        """
        Return a printable string representation of the call, useful for logging.
        """

        def arg_to_str(arg):
            if isinstance(arg, str):
                return f"'{arg}'"
            return str(arg)

        short_args = ", ".join(arg_to_str(a) for a in self.args)
        if len(short_args) > 200:
            short_args = short_args[:200] + "..."
        short_kwargs = ", ".join(f"{k}={arg_to_str(v)}" for k, v in self.kwargs.items())
        if len(short_kwargs) > 200:
            short_kwargs = short_kwargs[:200] + "..."
        args = ", ".join(a for a in [short_args, short_kwargs] if a)
        return f"{FunctionMap.get_readable_name(self.func_id)}({args})"


class Dispatcher(abc.ABC):
    """
    Abstract dispatcher to be inherited by all concrete dispatchers.
    For implementing a dispatcher you have to implement `_dispatch`, `srun` and `sbatch`.

    """

    @abc.abstractmethod
    def _dispatch(
        self, funcs: typing.Iterable[FunctionCall], options: SlurmOptions
    ) -> int:
        """
        Define how to dispatch a number of function calls.
        :param funcs: The function calls to be dispatched.
        :param options: The slurm options to be used.
        :return: The job id. Use -1 if not applicable (e.g., because buffered)
        """

    @abc.abstractmethod
    def srun(
        self,
        command: str,
        conf: typing.Optional[SlurmOptions] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ) -> int:
        """
        Define how you want to execute an `srun` command. This command is directly
        executed and only terminates after completion.
        :param command: A system command, e.g. `echo hello world > foobar.txt`.
        :param conf: The slurm configuration.
        :param simple_slurm_kwargs: Additional options for simple_slurm.
        :return: Job id
        """

    @abc.abstractmethod
    def sbatch(
        self,
        command: str,
        conf: typing.Optional[SlurmOptions] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ) -> int:
        """
        Define how you want to execute an `sbatch` command. The command is scheduled
        and the function return immediately.
        :param command: A system command, e.g. `echo hello world > foobar.txt`.
        :param conf: The slurm configuration.
        :param simple_slurm_kwargs: Additional options for simple_slurm.
        :return: Job id.
        """

    def _log_dispatch(self, funcs: typing.List[FunctionCall], options: SlurmOptions):
        if len(funcs) == 1:
            logging.getLogger("slurminade").info(
                f"Dispatching task with options {options}: {funcs[0]}"
            )
        else:
            logging.getLogger("slurminade").info(
                f"Dispatching task consisting of {len(funcs)} function calls with options {options}: {', '.join([str(f) for f in funcs])}"
            )

    def __call__(
        self,
        funcs: typing.Union[FunctionCall, typing.Iterable[FunctionCall]],
        options: SlurmOptions,
    ) -> int:
        """
        Dispatches a function call or a number of function calls.
        :param funcs: The function calls to be distributed.
        :param options: The slurm options to be used.
        :return: Job id.
        """
        if isinstance(funcs, FunctionCall):
            funcs = [funcs]
        funcs = list(funcs)
        self._log_dispatch(funcs, options)
        return self._dispatch(funcs, options)

    def is_sequential(self):
        """
        Return true if the dispatcher works sequential. In this case, the dependencies
        are trivially fulfilled. Slurm does not work sequentially, because this
        would destroy its purpose. In some cases however, you do not want to use
        slurm for compatibility reasons, without changing the script. In these cases,
        this function tells slurminade not to be too strict about dependencies.
        :return: True is tasks are executed sequentially, false if not.
        """
        return False


class TestDispatcher(Dispatcher):
    """
    A dummy dispatcher that just prints the output. Primarily for debugging and testing.
    """

    def __init__(self):
        super().__init__()
        self.calls = []
        self.sbatches = []
        self.sruns = []
        self.max_arg_length = DEFAULT_MAX_ARG_LENGTH

    def _dispatch(
        self, funcs: typing.Iterable[FunctionCall], options: SlurmOptions
    ) -> int:
        dispatch_guard()
        funcs = list(funcs)
        command = create_slurminade_command(funcs, self.max_arg_length)
        logging.getLogger("slurminade").info(command)
        self.calls.append(funcs)
        self._cleanup(command)
        return -1

    def _cleanup(self, command):
        args = shlex.split(command)
        if args[-2] != "temp":
            return
        filename = args[-1]
        if os.path.exists(filename):
            os.remove(filename)

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        dispatch_guard()
        self.sruns.append(command)
        print("SRUN", command)

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        dispatch_guard()
        self.sbatches.append(command)
        print("SBATCH", command)

    def is_sequential(self):
        return True


class SlurmDispatcher(Dispatcher):
    """
    The most important dispatcher: Distributing function calls to slurm.
    """

    def __init__(self):
        super().__init__()
        if not shutil.which("sbatch"):
            msg = "Slurm could not be found."
            raise RuntimeError(msg)
        self.max_arg_length = DEFAULT_MAX_ARG_LENGTH

    def _create_slurm_api(self, special_slurm_opts):
        conf = _get_conf(special_slurm_opts)
        slurm = simple_slurm.Slurm(**conf)
        return slurm

    def _job_name(self, funcs: typing.List[FunctionCall]) -> str:
        func_names = list({FunctionMap.get_readable_name(f.func_id) for f in funcs})
        if len(funcs) == 1:
            return f"slurminade:{func_names[0]}"
        else:
            return f"slurminade[batch]:{func_names[0]}..."

    def _dispatch(
        self, funcs: typing.Iterable[FunctionCall], options: SlurmOptions
    ) -> int:
        dispatch_guard()
        if "job_name" not in options:
            funcs = list(funcs)
            # This is complicated to prevent warnings about the type
            options = SlurmOptions(**options.as_dict())
            options["job_name"] = self._job_name(funcs)
        slurm = self._create_slurm_api(options)
        command = create_slurminade_command(funcs, self.max_arg_length)
        logging.getLogger("slurminade").debug(command)
        return slurm.sbatch(command)

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        dispatch_guard()
        conf = _get_conf(conf)
        slurm = simple_slurm.Slurm(**conf)
        logging.getLogger("slurminade").debug("SBATCH %s", command)
        if simple_slurm_kwargs:
            return slurm.sbatch(command, **simple_slurm_kwargs)
        else:
            return slurm.sbatch(command)

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        dispatch_guard()
        conf = _get_conf(conf)
        slurm = simple_slurm.Slurm(**conf)
        logging.getLogger("slurminade").debug("SRUN %s", command)
        if simple_slurm_kwargs:
            return slurm.srun(command, **simple_slurm_kwargs)
        else:
            return slurm.srun(command)


class SubprocessDispatcher(Dispatcher):
    """
    A dispatcher for debugging that distributes function calls using subprocesses.
    Thus, it uses the same serialization mechanisms, but without a slurm dependency.
    Completely useless for productive purposes. Use `DirectCallDispatcher` if you
    don't want to use slurm.
    Despite using subprocesses, it does not parallelize but works sequential.
    """

    def __init__(self):
        super().__init__()
        self.max_arg_length = DEFAULT_MAX_ARG_LENGTH

    def _dispatch(
        self, funcs: typing.Iterable[FunctionCall], options: SlurmOptions
    ) -> int:
        dispatch_guard()
        command = create_slurminade_command(funcs, self.max_arg_length)
        os.system(command)
        return -1

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        dispatch_guard()
        logging.getLogger("slurminade").debug("SRUN %s", command)
        subprocess.run(command, check=True)

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        self.srun(command)

    def is_sequential(self):
        return True


class DirectCallDispatcher(Dispatcher):
    """
    A dispatcher that calls functions as if we would not use slurminade.
    This allows compatibility of scripts also on computers not integrated into
    the slurm network.
    """

    def _dispatch(
        self, funcs: typing.Iterable[FunctionCall], options: SlurmOptions
    ) -> int:
        dispatch_guard()
        for func in funcs:
            FunctionMap.call(func.func_id, func.args, func.kwargs)
        return -1

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        dispatch_guard()
        subprocess.run(command, check=True)

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        self.srun(command)

    def is_sequential(self):
        return True


def create_slurminade_command(
    funcs: typing.Iterable[FunctionCall], max_arg_length: int
) -> str:
    """
    Creates a terminal command that calls the Python module `slurminade.execute` with the
    provided function calls as an argument. If the total length of the function calls
    exceeds the maximum allowed length of a command line argument, a temporary file is
    created to pass the function calls instead.
    :param funcs: The function calls to be dispatched.
    :param max_arg_length: The maximum allowed length of a command line argument.
    :returns: A string representing the command to be executed in the terminal.
    """
    command = f"{sys.executable} -m slurminade.execute --root {shlex.quote(get_entry_point())}"

    # Serialize function calls as JSON
    json_calls = json.dumps([f.to_json() for f in funcs])
    serialized_calls = shlex.quote(json_calls)

    if len(serialized_calls) > max_arg_length:
        # The argument is too long, create temporary file for the JSON
        fd, filename = mkstemp(prefix="slurminade_", suffix=".json", text=True, dir=".")
        logging.getLogger("slurminade").info(
            f"Long function calls. Serializing function calls to temporary file {filename}"
        )
        with os.fdopen(fd, "w") as f:
            f.write(json_calls)
        command += f" --fromfile {filename}"
    else:
        command += f" --calls {serialized_calls}"
    return command


# The current dispatcher. Use with `get_dispatcher` and `set_dispatcher`.
__dispatcher: typing.Optional[Dispatcher] = None


def get_dispatcher() -> Dispatcher:
    """
    Returns the current dispatcher. Creates a dispatcher if none is available.
    First tries to create the slurm-dispatcher (as this is the primary purpose of
    slurminade). If no slurm-environment can be found, it creates a DirectCallDispatcher
    to allow compatibility.
    :return: The dispatcher.
    """
    global __dispatcher
    if __dispatcher is None:
        try:
            __dispatcher = SlurmDispatcher()
        except RuntimeError as re:
            logging.getLogger("slurminade").warning(str(re))
            logging.getLogger("slurminade").warning("Using direct calls.")
            __dispatcher = DirectCallDispatcher()
    return __dispatcher


def set_dispatcher(dispatcher: Dispatcher) -> None:
    """
    Replaces the dispatcher. Can be used to enforce a specific dispatcher.
    :param dispatcher: The dispatcher to be used.
    :return: None
    """
    global __dispatcher
    __dispatcher = dispatcher
    assert dispatcher == get_dispatcher()


def dispatch(
    funcs: typing.Union[FunctionCall, typing.Iterable[FunctionCall]],
    options: SlurmOptions,
) -> int:
    """
    Distribute function calls with the current dispatcher.
    :param funcs: The functions calls to be distributed.
    :param options: The slurm options to be used.
    :return: The job id.
    """
    funcs = list(funcs) if not isinstance(funcs, FunctionCall) else [funcs]
    for func in funcs:
        if not FunctionMap.check_id(func.func_id):
            msg = f"Function '{func.func_id}' cannot be called from the given entry point."
            raise KeyError(msg)
    return get_dispatcher()(funcs, options)


def srun(
    command: str,
    conf: typing.Union[SlurmOptions, typing.Dict, None] = None,
    simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
) -> int:
    """
    Call srun with the current dispatcher. This command is directly
    executed and only terminates after completion.
    :param command: A system command, e.g. `echo hello world > foobar.txt`.
    :param conf: The slurm configuration.
    :param simple_slurm_kwargs: Additional options for simple_slurm.
    :return: Job id
    """
    if not isinstance(conf, SlurmOptions):
        if conf is None:
            conf = {}
        conf = SlurmOptions(**conf)
    return get_dispatcher().srun(command, conf, simple_slurm_kwargs)


def sbatch(
    command: str,
    conf: typing.Union[SlurmOptions, typing.Dict, None] = None,
    simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
) -> int:
    """
    The command is scheduled and the function returns immediately.
    :param command: A system command, e.g. `echo hello world > foobar.txt`.
    :param conf: The slurm configuration.
    :param simple_slurm_kwargs: Additional options for simple_slurm.
    :return: Job id.
    """
    if not isinstance(conf, SlurmOptions):
        if conf is None:
            conf = {}
        conf = SlurmOptions(**conf)
    return get_dispatcher().sbatch(command, conf, simple_slurm_kwargs)



import os.path
import unittest

import slurminade
from slurminade.function import SlurmFunction

f_file = "./f_test_file.txt"
g_file = "./g_test_file.txt"


@slurminade.slurmify()
def f():
    with open(f_file, "w") as file:
        file.write("test")


def delete_f():
    if os.path.exists(f_file):
        os.remove(f_file)


@slurminade.slurmify()
def g(x, y):
    with open(g_file, "w") as file:
        file.write(f"{x}:{y}")


def delete_g():
    if os.path.exists(g_file):
        os.remove(g_file)


class TestLocal(unittest.TestCase):
    def test_1(self):
        delete_f()
        SlurmFunction.call(f.func_id)
        assert os.path.exists(f_file)
        delete_f()

    def test_2(self):
        delete_g()
        SlurmFunction.call(g.func_id, x="a", y=2)
        assert os.path.exists(g_file)
        with open(g_file) as file:
            assert file.readline() == "a:2"
        delete_g()



"""
Contains code for bundling function calls together.
"""
import logging
import typing
from collections import defaultdict

from .dispatcher import (
    Dispatcher,
    FunctionCall,
    get_dispatcher,
    set_dispatcher,
)
from .function import SlurmFunction
from .guard import BatchGuard
from .options import SlurmOptions


class TaskBuffer:
    """
    A simple container to buffer all the tasks by their options.
    We can only bundle tasks with the same slurm options.
    """

    def __init__(self):
        self._tasks = defaultdict(list)

    def add(self, task: FunctionCall, options: SlurmOptions) -> int:
        self._tasks[options].append(task)
        return len(self._tasks[options])

    def items(self):
        for opt, tasks in self._tasks.items():
            if tasks:
                yield opt, tasks

    def get(self, options: SlurmOptions) -> typing.List[FunctionCall]:
        return self._tasks[options]

    def clear(self):
        self._tasks.clear()


class Batch(Dispatcher):
    """
    The logic to buffer the function calls. It wraps the original dispatcher.

    You can use::

        with slurminade.Batch(max_size=20) as batch:  # automatically bundles up to 20 tasks
            # run 100x f
            for i in range(100):
                f.distribute(i)

    to automatically bundle up to 20 tasks and distribute them.
    """

    def __init__(self, max_size: int):
        """
        :param max_size: Bundle up to this many calls.
        """
        super().__init__()
        self.max_size = max_size
        self.subdispatcher = get_dispatcher()
        self._tasks = TaskBuffer()
        self._batch_guard = BatchGuard()

    def flush(self, options: typing.Optional[SlurmOptions] = None) -> typing.List[int]:
        """
        Distribute all buffered tasks. Return the job ids used.
        This method is called automatically when the context is exited.
        However, you may want to call it manually to get the job ids,
        for example to use them for dependency management with ``wait_for``.
        :param options: Only flush tasks with specific options.
        :return: A list of job ids.
        """
        job_ids = []
        if options is None:
            for opt, tasks in self._tasks.items():
                while tasks:
                    job_id = self.subdispatcher(tasks[: self.max_size], opt)
                    job_ids.append(job_id)
                    tasks = tasks[self.max_size :]

        else:
            tasks = self._tasks.get(options)
            self._batch_guard.report_flush(len(tasks))
            while len(tasks) > self.max_size:
                job_id = self.subdispatcher(tasks[: self.max_size], options)
                job_ids.append(job_id)
                tasks = tasks[: self.max_size]
        self._tasks.clear()
        return job_ids

    def add(self, func: SlurmFunction, *args, **kwargs):
        """
        You can also add a task using `add` instead of `distribute`.
        :param func: The function to call
        :param args: The positional arguments
        :param kwargs: The keywords arguments.
        :return: None
        """
        self._dispatch(
            [FunctionCall(func.func_id, args, kwargs)], func.special_slurm_opts
        )

    def _dispatch(
        self, funcs: typing.Iterable[FunctionCall], options: SlurmOptions
    ) -> int:
        for func in funcs:
            self._tasks.add(func, options)
        return -1

    def srun(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        return self.subdispatcher.srun(command, conf, simple_slurm_kwargs)

    def sbatch(
        self,
        command: str,
        conf: typing.Optional[typing.Dict] = None,
        simple_slurm_kwargs: typing.Optional[typing.Dict] = None,
    ):
        return self.subdispatcher.sbatch(command, conf, simple_slurm_kwargs)

    def __enter__(self):
        self.subdispatcher = get_dispatcher()
        set_dispatcher(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            logging.getLogger("slurminade").error("Aborted due to exception.")
            return
        self.flush()
        set_dispatcher(self.subdispatcher)

    def _log_dispatch(self, funcs: typing.List[FunctionCall], options: SlurmOptions):
        if len(funcs) == 1:
            logging.getLogger("slurminade").info(
                f"Adding task to batch with options {options}: {funcs[0]}"
            )
        else:
            logging.getLogger("slurminade").info(
                f"Adding {len(funcs)} tasks to batch with options {options}: {', '.join([str(f) for f in funcs])}"
            )

    def __del__(self):
        self.flush()

    def is_sequential(self):
        return self.subdispatcher.is_sequential()



"""
This file saves the default configuration for slurm.
"""

import json
import os.path
import typing
from pathlib import Path

CONFIG_NAME = ".slurminade_default.json"

__default_conf: typing.Dict = {}


def _load_conf(path):
    try:
        if os.path.isfile(path):
            with open(path) as f:
                return json.load(f)
        else:
            return {}
    except Exception as e:
        print(f"slurminade could not open default configuration {path}!\n{e!s}")
    return {}


def update_default_configuration(conf=None, **kwargs):
    if conf:
        __default_conf.update(conf)
    if kwargs:
        __default_conf.update(kwargs)


def _load_default_conf():
    path = os.path.join(Path.home(), CONFIG_NAME)
    update_default_configuration(_load_conf(path))
    if "XDG_CONFIG_HOME" in os.environ:
        path = os.path.join(os.environ["XDG_CONFIG_HOME"], "slurminade", CONFIG_NAME)
        update_default_configuration(_load_conf(path))
    update_default_configuration(_load_conf(CONFIG_NAME))


_load_default_conf()


def set_default_configuration(conf=None, **kwargs):
    __default_conf = {}
    update_default_configuration(conf, **kwargs)


def _get_conf(conf=None):
    conf = conf if conf else {}
    conf_ = __default_conf.copy()
    conf_.update(conf)
    return conf_
